import os
import uuid

from anvil import CONFIG
from anvil.lib.format_versions import *
from anvil.lib.lib import File


class JsonSchemes:
    """A class used to read and write to the json_schemes.json file."""

    @staticmethod
    def pack_name_lang(name, description):
        return [f"pack.name={name}", f"pack.description={description}"]

    @staticmethod
    def skin_pack_name_lang(name, display_name):
        return [f"skinpack.{name}={display_name}"]

    @staticmethod
    def manifest_bp(version):
        m = {
            "format_version": 2,
            "header": {
                "description": "pack.description",
                "name": "pack.name",
                "uuid": CONFIG._BP_UUID[0],
                "version": version,
                "min_engine_version": [int(i) for i in MANIFEST_BUILD.split(".")],
            },
            "modules": [{"type": "data", "uuid": str(uuid.uuid4()), "version": version}],
            "dependencies": [{"uuid": CONFIG._RP_UUID[0], "version": version}],
            "metadata": {"authors": [CONFIG.COMPANY]},
        }
        if CONFIG._SCRIPT_API:
            m["modules"].append(
                {
                    "uuid": str(uuid.uuid4()),
                    "version": version,
                    "type": "script",
                    "language": "javascript",
                    "entry": "scripts/main.js",
                }
            )
            m["dependencies"].append(
                {
                    "module_name": "@minecraft/server",
                    "version": MODULE_MINECRAFT_SERVER,
                }
            )
            if CONFIG._SCRIPT_UI:
                m["dependencies"].append(
                    {
                        "module_name": "@minecraft/server-ui",
                        "version": MODULE_MINECRAFT_SERVER_UI,
                    }
                )
        if CONFIG._TARGET == "addon":
            m["header"]["pack_scope"] = "world"
            m["metadata"]["product_type"] = "addon"
        return m

    @staticmethod
    def manifest_rp(version):
        m = {
            "format_version": 2,
            "header": {
                "description": "pack.description",
                "name": "pack.name",
                "uuid": CONFIG._RP_UUID[0],
                "version": version,
                "min_engine_version": [int(i) for i in MANIFEST_BUILD.split(".")],
            },
            "modules": [
                {
                    "type": "resources",
                    "uuid": str(uuid.uuid4()),
                    "version": version,
                }
            ],
            "dependencies": [{"uuid": CONFIG._BP_UUID[0], "version": version}],
            "metadata": {"authors": [CONFIG.COMPANY]},
        }
        if CONFIG._PBR:
            m.update({"capabilities": ["pbr"]})
        if CONFIG._TARGET == "addon":
            m["header"]["pack_scope"] = "world"
            m["metadata"]["product_type"] = "addon"
        return m

    @staticmethod
    def manifest_world(version):
        m = {
            "format_version": 2,
            "header": {
                "name": "pack.name",
                "description": "pack.description",
                "version": version,
                "uuid": CONFIG._PACK_UUID,
                # "platform_locked": False,
                "lock_template_options": True,
                "base_game_version": [int(i) for i in MANIFEST_BUILD.split(".")],
            },
            "modules": [
                {
                    "type": "world_template",
                    "uuid": str(uuid.uuid4()),
                    "version": version,
                }
            ],
            "metadata": {"authors": [CONFIG.COMPANY]},
        }

        if CONFIG._RANDOM_SEED:
            m["header"]["allow_random_seed"] = True

        return m

    @staticmethod
    def world_packs(version, pack_id):
        return [{"pack_id": i, "version": version} for i in pack_id]

    @staticmethod
    def code_workspace(name, path1, path2):
        return {
            "folders": [
                {"name": name, "path": os.path.join(path1, path2)},
            ]
        }

    @staticmethod
    def packagejson(project_name, version, description, author):
        return {
            "name": project_name,
            "version": version,
            "description": description,
            "main": "scripts/main.js",
            "scripts": {"test": 'echo "Error: no test specified" && exit 1'},
            "keywords": [],
            "author": author,
            "license": "ISC",
        }

    @staticmethod
    def skins_json(serialize_name):
        return {
            "serialize_name": serialize_name,
            "localization_name": serialize_name,
            "skins": [],
        }

    @staticmethod
    def skin_json(filename: str, is_slim: bool, free: bool):
        return {
            "localization_name": filename,
            "geometry": f"geometry.humanoid.{ 'customSlim' if is_slim else 'custom'}",
            "texture": f"{filename}.png",
            "type": "free" if free else "paid",
        }

    @staticmethod
    def manifest_skins(version):
        return {
            "format_version": 1,
            "header": {
                "name": "pack.name",
                "uuid": str(uuid.uuid4()),
                "version": version,
            },
            "modules": [
                {
                    "type": "skin_pack",
                    "uuid": str(uuid.uuid4()),
                    "version": version,
                }
            ],
        }

    @staticmethod
    def description(namespace, identifier):
        return {"description": {"identifier": f"{namespace}:{identifier}"}}

    @staticmethod
    def item_texture(resource_pack_name):
        return {
            "resource_pack_name": resource_pack_name,
            "texture_name": "atlas.items",
            "texture_data": {},
        }

    @staticmethod
    def sound_definitions():
        return {
            "format_version": SOUND_DEFINITIONS_VERSION,
            "sound_definitions": {},
        }

    @staticmethod
    def music_definitions():
        return {}

    @staticmethod
    def sound(name, category):
        return {name: {"category": category, "sounds": []}}

    @staticmethod
    def materials():
        return {"materials": {"version": MATERIALS_VERSION}}

    @staticmethod
    def languages():
        return [
            "en_US",
            "en_GB",
            "de_DE",
            "es_ES",
            "es_MX",
            "fr_FR",
            "fr_CA",
            "it_IT",
            "pt_BR",
            "pt_PT",
            "ru_RU",
            "zh_CN",
            "zh_TW",
            "nl_NL",
            "bg_BG",
            "cs_CZ",
            "da_DK",
            "el_GR",
            "fi_FI",
            "hu_HU",
            "id_ID",
            "nb_NO",
            "pl_PL",
            "sk_SK",
            "sv_SE",
            "tr_TR",
            "uk_UA",
        ]

    @staticmethod
    def client_description():
        return {
            "materials": {"default": "entity_alphatest"},
            "scripts": {"pre_animation": [], "initialize": [], "animate": []},
            "textures": {},
            "geometry": {},
            "particle_effects": {},
            "sound_effects": {},
            "render_controllers": [],
        }

    @staticmethod
    def client_entity():
        return {
            "format_version": ENTITY_CLIENT_VERSION,
            "minecraft:client_entity": {},
        }

    @staticmethod
    def server_entity():
        return {
            "format_version": ENTITY_SERVER_VERSION,
            "minecraft:entity": {
                "component_groups": {},
                "components": {},
                "events": {},
            },
        }

    @staticmethod
    def bp_animations():
        return {"format_version": BP_ANIMATION_VERSION, "animations": {}}

    @staticmethod
    def bp_animation(animation_name, part, animation_type, loop):
        return {
            f"animation.{animation_name}.{part}.{animation_type}": {
                "loop": loop,
                "timeline": {},
            }
        }

    @staticmethod
    def rp_animations():
        return {"format_version": RP_ANIMATION_VERSION, "animations": {}}

    @staticmethod
    def animation_controller_state(state):
        return {
            state: {
                "on_entry": [],
                "on_exit": [],
                "animations": [],
                "transitions": [],
            }
        }

    @staticmethod
    def animation_controller(controller_name, part, animation_type):
        return {
            f"controller.animation.{controller_name}.{part}.{animation_type}": {
                "initial_state": "default",
                "states": {},
            }
        }

    @staticmethod
    def animation_controllers():
        return {
            "format_version": ANIMATION_CONTROLLERS_VERSION,
            "animation_controllers": {},
        }

    @staticmethod
    def geometry(model_name: str, texture_size: list[int], visible_box: list[int], visible_offset: list[int]):
        return {
            "format_version": GEOMETRY_VERSION,
            "minecraft:geometry": [
                {
                    "description": {
                        "identifier": f"geometry.{CONFIG.NAMESPACE}.{model_name}",
                        "texture_width": texture_size[0],
                        "texture_height": texture_size[1],
                        "visible_bounds_width": visible_box[0],
                        "visible_bounds_height": visible_box[1],
                        "visible_bounds_offset": visible_offset,
                    },
                    "bones": [],
                }
            ],
        }

    @staticmethod
    def render_controller(controller_name, part, controller_type):
        return {
            f"controller.render.{controller_name}.{part}.{controller_type}": {
                "arrays": {"textures": {}, "geometries": {}},
                "materials": [],
                "geometry": {},
                "textures": [],
                "part_visibility": [],
            }
        }

    @staticmethod
    def render_controllers():
        return {
            "format_version": RENDER_CONTROLLER_VERSION,
            "render_controllers": {},
        }

    @staticmethod
    def attachable():
        return {"format_version": ENTITY_CLIENT_VERSION, "minecraft:attachable": {}}

    @staticmethod
    def spawn_rules():
        return {
            "format_version": SPAWN_RULES_VERSION,
            "minecraft:spawn_rules": {"conditions": []},
        }

    @staticmethod
    def server_block():
        return {
            "format_version": BLOCK_SERVER_VERSION,
            "minecraft:block": {
                "description": {},
                "components": {},
                "permutations": [],
            },
        }

    @staticmethod
    def terrain_texture(resource_pack_name):
        return {
            "num_mip_levels": 4,
            "padding": 8,
            "resource_pack_name": resource_pack_name,
            "texture_data": {},
            "texture_name": "atlas.terrain",
        }

    @staticmethod
    def flipbook_textures():
        return []

    @staticmethod
    def font(font_name, font_file):
        return {
            "version": 1,
            "fonts": [
                {
                    "font_format": "ttf",
                    "font_name": font_name,
                    "version": 1,
                    "font_file": f"font/{font_file}",
                    "lowPerformanceCompatible": False,
                }
            ],
        }

    @staticmethod
    def fog():
        return {
            "format_version": FOG_VERSION,
            "minecraft:fog_settings": {
                "distance": {},
                "volumetric": {},
            },
        }

    @staticmethod
    def dialogues():
        return {
            "format_version": DIALOGUE_VERSION,
            "minecraft:npc_dialogue": {"scenes": []},
        }

    @staticmethod
    def dialogue_scene(scene_tag, npc_name, text, on_open_commands, on_close_commands, buttons):
        return {
            "scene_tag": scene_tag,
            "npc_name": npc_name,
            "text": text,
            "on_open_commands": on_open_commands,
            "on_close_commands": on_close_commands,
            "buttons": buttons,
        }

    @staticmethod
    def dialogue_button(name, commands):
        return {"name": name, "commands": commands}

    @staticmethod
    def server_item():
        return {
            "format_version": ITEM_SERVER_VERSION,
            "minecraft:item": {"components": {}},
        }

    @staticmethod
    def camera_preset(namespace, name, inherit_from):
        return {
            "format_version": CAMERA_PRESET_VERSION,
            "minecraft:camera_preset": {
                "identifier": f"{namespace}:{name}",
                "inherit_from": inherit_from,
            },
        }

    @staticmethod
    def tsconfig(pascal_project_name):
        return {
            "compilerOptions": {
                "target": "ESNext",
                "module": "es2020",
                "declaration": False,
                "outDir": f"behavior_packs/BP_{pascal_project_name}/scripts",
                "strict": True,
                "pretty": True,
                "esModuleInterop": True,
                "moduleResolution": "Node",
                "resolveJsonModule": True,
                "forceConsistentCasingInFileNames": True,
                "lib": [
                    "ESNext",
                    "dom",
                ],
            },
            "include": ["assets/javascript/**/*"],
            "exclude": ["node_modules"],
        }

    @staticmethod
    def vscode(pascal_project_name):
        return {
            "version": "0.3.0",
            "configurations": [
                {
                    "type": "minecraft-js",
                    "request": "attach",
                    "name": "Wait for Minecraft Debug Connections",
                    "mode": "listen",
                    "localRoot": f"${{workspaceFolder}}/behavior_packs/BP_{pascal_project_name}/scripts",
                    "port": 19144,
                }
            ],
        }

    @staticmethod
    def blocks():
        return {
            "format_version": [int(i) for i in BLOCK_JSON_FORMAT_VERSION.split(".")],
        }

    @staticmethod
    def sounds():
        return {
            "individual_event_sounds": {},
            "block_sounds": {},
            "entity_sounds": {"entities": {}},
            "interactive_sounds": {},
        }

    @staticmethod
    def directional_lights():
        return {
            "format_version": GLOBAL_LIGHTING.split("."),
            "directional_lights": {},
            "pbr": {},
        }

    @staticmethod
    def atmospherics():
        return {
            "horizon_blend_stops": {},
        }

    @staticmethod
    def loot_table():
        return {"pools": []}

    @staticmethod
    def smelting_recipe(namespace: str, name: str, tags: list[str]):
        return {
            "format_version": "1.16.0",
            "minecraft:recipe_furnace": {
                "tags": tags,
                "description": {"identifier": f"{namespace}:{name}"},
            },
        }

    @staticmethod
    def smithing_table_recipe(namespace: str, name: str, tags: list[str]):
        return {
            "format_version": "1.17.0",
            "minecraft:recipe_smithing_transform": {
                "description": {"identifier": f"{namespace}:{name}"},
                "tags": tags,
            },
        }

    @staticmethod
    def smithing_table_trim_recipe(namespace: str, name: str, tags: list[str]):
        return {
            "format_version": "1.17.0",
            "minecraft:recipe_smithing_trim": {
                "description": {"identifier": f"{namespace}:{name}"},
                "tags": tags,
            },
        }

    @staticmethod
    def shapeless_crafting_recipe(namespace: str, name: str, tags: list[str]):
        return {
            "format_version": "1.12",
            "minecraft:recipe_shapeless": {
                "description": {"identifier": f"{namespace}:{name}"},
                "tags": tags,
                "ingredients": [],
            },
        }

    @staticmethod
    def shaped_crafting_recipe(namespace: str, name: str, assume_symmetry: bool, tags: list[str]):
        return {
            "format_version": "1.12",
            "minecraft:recipe_shaped": {
                "description": {"identifier": f"{namespace}:{name}"},
                "tags": tags,
                "assume_symmetry": assume_symmetry,
                "pattern": [],
                "key": {},
            },
        }

    @staticmethod
    def tsconstants(namespace: str, project_name: str):
        file = []
        file.append(f'export const NAMESPACE = "{namespace}"')
        file.append(f'export const PROJECT_NAME = "{project_name}"')

        return "\n".join(file)


class AddonObject:
    """
    An object representing an addon with functionality to modify its content, queue it for processing, and export it.

    Attributes:
        name (str): The name of the addon object.
    """

    _extension = ".json"
    _path = ""

    def __init__(self, name: str) -> None:
        """
        Constructs all the necessary attributes for the AddonObject object.

        Args:
            name (str): The name of the addon object.
            path (str): The path of the addon object.
        """
        self._shorten = True
        self._name = name
        self._content = {}
        self._directory = ""
        CONFIG.Logger.object_initiated(self._name)

    @property
    def do_not_shorten(self):
        """
        Setter property that disables shortening of `dict` when exporting.
        """
        self._shorten = False

    def content(self, content):
        """
        Sets the content of the addon object and returns the object.

        Args:
            content (any): The content to be set for the addon object.

        Returns:
            self: The instance of the current AddonObject.
        """
        self._content = content
        return self

    def queue(self, directory: str | None = None):
        """
        Queues the addon object for processing and logs the event.

        Args:
            directory (str, optional): The directory of the addon object.

        Returns:
            self: The instance of the current AddonObject.
        """
        from anvil import ANVIL

        self._directory = directory if not directory is None else ""
        self._path = os.path.join(self._path, self._directory)
        CONFIG.Logger.object_queued(self._name)
        ANVIL._queue(self)
        return self

    def _export(self):
        """
        Exports the addon object after potentially shortening its content and replacing backslashes.
        Logs the event and writes the object to a file.
        """

        def _shorten_dict(d):
            if isinstance(d, dict):
                return {
                    k: (v if v != {"do_not_shorten": True} else {})
                    for k, v in ((k, _shorten_dict(v)) for k, v in d.items())
                    if (v != {} and v != []) or str(k).startswith("minecraft:") or v == {"do_not_shorten": True}
                }

            elif isinstance(d, list):
                return [v for v in map(_shorten_dict, d) if v != []]

            return d

        def _replace_backslashes(obj):
            if isinstance(obj, str):
                return obj.replace('"/n"', '"\\n"').replace("/n", "\n").replace("\\", "/")
            elif isinstance(obj, list):
                return [_replace_backslashes(item) for item in obj]
            elif isinstance(obj, dict):
                return {key: _replace_backslashes(value) for key, value in obj.items()}
            else:
                return obj

        path = self._path.removeprefix(CONFIG.RP_PATH).removeprefix(CONFIG.BP_PATH)
        path = os.path.join(path, f"{self._name}{self._extension}")
        if len(path) > 80:
            CONFIG.Logger.path_length_error(path)

        if self._shorten and type(self._content) is dict:
            self._content = _shorten_dict(self._content)

        self._content = _replace_backslashes(self._content)
        CONFIG.Logger.object_exported(self._name)
        File(f"{self._name}{self._extension}", self._content, self._path, "w")


class MinecraftDescription:
    """Handles Minecraft descriptions.

    Attributes:
        name (str): The name of the Minecraft object.
        is_vanilla (bool, optional): If the object is from vanilla Minecraft. Defaults to False.
    """

    def _validate_name(self, name: str):
        """Validates the name of the Minecraft object.

        Args:
            name (str): The name of the Minecraft object.
        """
        if ":" in name:
            CONFIG.Logger.namespaces_not_allowed(name)

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Initializes a MinecraftDescription instance.

        Args:
            name (str): The name of the Minecraft object.
            is_vanilla (bool, optional): If the object is from vanilla Minecraft. Defaults to False.
        """
        self._validate_name(name)
        self._name = name
        self._is_vanilla = is_vanilla
        self._namespace = "minecraft" if is_vanilla else CONFIG.NAMESPACE
        self._description: dict = JsonSchemes.description(self._namespace, self._name)

    @property
    def identifier(self) -> str:
        """Formulates the identifier for the Minecraft object.

        Returns:
            str: The identifier in the format 'namespace_format:name'
        """
        return f"{self._namespace}:{self._name}"

    @property
    def name(self) -> str:
        """Returns the name of the Minecraft object.

        Returns:
            str: The name
        """
        return self._name

    @property
    def is_vanilla(self) -> bool:
        """Returns if the Minecraft object is from vanilla Minecraft.

        Returns:
            bool: If the object is from vanilla Minecraft
        """
        return self._is_vanilla

    @property
    def to_dict(self) -> dict:
        """Returns the description of the Minecraft object.

        Returns:
            dict: The description
        """
        return self._description



import os
from enum import StrEnum

from anvil import CONFIG
from anvil.api.enums import EntitySoundEvent, MusicCategory, SoundCategory
from anvil.api.types import Identifier
from anvil.lib.lib import CopyFiles, FileExists
from anvil.lib.reports import ReportType
from anvil.lib.schemas import AddonObject, JsonSchemes


class SoundDescription:
    def __init__(
        self,
        sound_reference: str,
        category: SoundCategory,
        use_legacy_max_distance: bool = False,
        max_distance: float = 0,
        min_distance: float = 9999,
    ) -> None:
        """Initializes a SoundDescription instance.

        Args:
            sound_reference (str): The definition of the sound.
            category (SoundCategory): The category of the sound.
            use_legacy_max_distance (bool, optional): Use legacy max distance if True. Defaults to False.
            max_distance (int, optional): The maximum distance for the sound. Defaults to 0.
            min_distance (int, optional): The minimum distance for the sound. Defaults to 9999.
        """
        self._category = category.value
        self._sound_reference = sound_reference
        self._sound_definition = f"{CONFIG.NAMESPACE}:{sound_reference}"
        self._sound = JsonSchemes.sound(self._sound_definition, self._category)

        if use_legacy_max_distance != False:
            self._sound[self._sound_definition].update({"__use_legacy_max_distance": use_legacy_max_distance})
        if max_distance != 0:
            self._sound[self._sound_definition].update({"max_distance": max_distance})
        if min_distance != 9999:
            self._sound[self._sound_definition].update({"min_distance": min_distance})
        self._sounds = []

    def add_sound(
        self,
        sound_name,
        volume: int = 1,
        weight: int = 1,
        pitch: int = [1, 1],
        is_3d: bool = None,
        stream: bool = None,
        load_on_low_memory: bool = False,
    ):
        """Adds a sound to the SoundDescription instance.

        Args:
            sound_name (str): The name of the sound.
            volume (int, optional): The volume of the sound. Defaults to 1.
            weight (int, optional): The weight of the sound. Defaults to 1.
            pitch (list[int], optional): The pitch of the sound. Defaults to [1, 1].
            is_3d (bool, optional): If the sound is 3D. Defaults to None.
            stream (bool, optional): If the sound is streamed. Defaults to None.
            load_on_low_memory (bool, optional): If the sound is loaded on low memory. Defaults to False.
        """
        parent_dir = self._sound_reference.split(".")[0]
        if not FileExists(os.path.join("assets", "sounds", parent_dir, f"{sound_name}.ogg")):
            CONFIG.Logger.file_exist_error(f"{sound_name}.ogg", os.path.join("assets", "sounds", parent_dir))

        self._sound_name = sound_name
        splits = self._sound_definition.removeprefix(f"{CONFIG.NAMESPACE}:").split(".")
        self._path = CONFIG.NAMESPACE
        for i in range(len(splits) - 1):
            self._path = os.path.join(self._path, splits[i])
        sound = {"name": os.path.join("sounds", self._path, self._sound_name), "load_on_low_memory": load_on_low_memory}
        if pitch != [1, 1]:
            sound.update({"pitch": pitch})
        if not is_3d is None:
            sound.update({"is3D": is_3d})
        if not stream is None:
            sound.update({"stream": stream})
        if volume != 1:
            sound.update({"volume": volume})
        if weight != 1:
            sound.update({"weight": weight})
        self._sounds.append(sound)

        CONFIG.Report.add_report(ReportType.SOUND, col0=self._sound_definition, col1=self._sound_name, vanilla=False)
        return self

    @property
    def identifier(self):
        return self._sound_definition

    @property
    def _export(self):
        """Returns the sound description.

        Returns:
            dict: The sound description
        """
        parent_dir = self._sound_reference.split(".")[0]
        for sound in self._sounds:
            s = sound["name"].split("\\")[-1]
            CopyFiles(
                os.path.join("assets", "sounds", parent_dir),
                os.path.join(
                    CONFIG.RP_PATH,
                    "sounds",
                    self._path,
                ),
                f"{s}.ogg",
            )
        self._sound[self._sound_definition]["sounds"] = self._sounds
        return self._sound


class SoundDefinition(AddonObject):
    """Handles sound definitions.

    Attributes:
        _sounds (list[SoundDescription], optional): A list of sound descriptions. Defaults to empty list.
    """

    _path = os.path.join(CONFIG.RP_PATH, "sounds")

    def __init__(self) -> None:
        """Initializes a SoundDefinition instance."""
        super().__init__("sound_definitions")
        self.content(JsonSchemes.sound_definitions())
        self._sounds: list[SoundDescription] = []

    def sound_reference(
        self,
        sound_reference: str,
        category: SoundCategory,
        use_legacy_max_distance: bool = False,
        max_distance: float = 0,
        min_distance: float = 9999,
    ):
        """Defines a sound for the SoundDefinition instance.

        Args:
            sound_reference (str): The sound_reference of the sound.
            category (SoundCategory): The category of the sound.
            use_legacy_max_distance (bool, optional): Use legacy max distance if True. Defaults to False.
            max_distance (int, optional): The maximum distance for the sound. Defaults to 0.
            min_distance (int, optional): The minimum distance for the sound. Defaults to 9999.

        Returns:
            SoundDescription: The created sound description instance.
        """
        sound = SoundDescription(
            sound_reference,
            category,
            use_legacy_max_distance,
            max_distance,
            min_distance,
        )
        self._sounds.append(sound)
        return sound

    @property
    def queue(self):
        """Returns the queue for the sound definition.

        Returns:
            queue: The queue for the sound definition.
        """
        return super().queue("")

    def _export(self):
        """Returns the sound definition.

        Returns:
            dict: The sound definition.
        """
        for sound in self._sounds:
            self._content["sound_definitions"].update(sound._export)
        return super()._export()


class MusicDefinition(AddonObject):
    """Handles music definitions.

    Attributes:
        _extensions (dict): The file extensions for the music definition.
        _sounds (list[SoundDescription], optional): A list of sound descriptions. Defaults to empty list.
    """

    _path = os.path.join(CONFIG.RP_PATH, "sounds")

    def __init__(self) -> None:
        """Initializes a MusicDefinition instance."""
        super().__init__("music_definitions")
        self.content(JsonSchemes.music_definitions())

    def music_definition(self, music_category: MusicCategory, min_delay: int = 60, max_delay: int = 180):
        """Defines a music for the MusicDefinition instance.

        Args:
            music_category (MusicCategory): The category of the music.
            min_delay (int, optional): The minimum delay for the music. Defaults to 60.
            max_delay (int, optional): The maximum delay for the music. Defaults to 180.

        Returns:
            SoundDescription: The created sound description instance.
        """
        self._content.update(
            {
                music_category.value: {
                    "event_name": f"{CONFIG.NAMESPACE}:music.{music_category}",
                    "max_delay": max_delay,
                    "min_delay": min_delay,
                }
            }
        )

    @property
    def queue(self):
        """Returns the queue for the music definition.

        Returns:
            queue: The queue for the music definition.
        """
        return super().queue("")

    def _export(self):
        """Returns the music definition.

        Returns:
            dict: The music definition.
        """
        return super()._export()


class SoundEvent(AddonObject):
    """Handles sound events."""

    _path = os.path.join(CONFIG.RP_PATH)

    def __init__(self) -> None:
        """Initializes a SoundEvent instance."""
        super().__init__("sounds")
        self.content(JsonSchemes.sounds())

    def add_entity_event(
        self,
        entity_identifier: Identifier,
        sound_identifier: str,
        sound_event: EntitySoundEvent,
        volume: float = 1.0,
        pitch: tuple[float, float] = (0.8, 1.2),
    ):
        """Adds an entity event to the SoundEvent instance.

        Args:
            entity_identifier (Identifier): The identifier of the entity.
            sound_event (EntitySoundEvent): The sound event for the entity.
            sound_identifier (str): The identifier of the sound.
            volume (float, optional): The volume of the event. Defaults to 1.0.
            pitch (tuple[float, float], optional): The pitch of the event. Defaults to (0.8, 1.2).
        """

        if not entity_identifier in self._content["entity_sounds"]["entities"]:
            self._content["entity_sounds"]["entities"][entity_identifier] = {"volume": volume, "pitch": (0.8, 1.2), "events": {}}

        self._content["entity_sounds"]["entities"][entity_identifier]["events"][sound_event.value] = {
            "sound": f"{CONFIG.NAMESPACE}:{sound_identifier}"
        }

        if pitch != (0.8, 1.2):
            self._content["entity_sounds"]["entities"][entity_identifier]["events"][sound_event.value]["pitch"] = pitch

    @property
    def queue(self):
        """Returns the queue for the sound event.

        Returns:
            queue: The queue for the sound event.
        """
        return super().queue()



from collections import defaultdict
from enum import StrEnum


class ReportType(StrEnum):
    SOUND = "sound"
    ENTITY = "entity"
    ATTACHABLE = "attachable"
    ITEM = "item"
    BLOCK = "block"
    PARTICLE = "particle"


class ReportCollector:
    def __init__(self) -> None:
        """A dictionary where keys are ReportType and values are `list[defaultdict]`.
        These inner `defaultdict` will automatically create a key if it doesn't exist
        and initialize it to a dictionary of empty sets.

        The dictionary is later used to generate a report of what have been added to the pack.
        """
        self.dict = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))

    def add_headers(self):
        self.add_report(ReportType.SOUND, col0="Sound Identifier", col1="Sounds", vanilla=False)
        self.add_report(ReportType.ENTITY, col0="Entity Name", col1="Entity Identifier", col2="Entity Events", vanilla=False)
        self.add_report(ReportType.ATTACHABLE, col0="Attachable Name", col1="Attachable Identifier", vanilla=False)
        self.add_report(ReportType.ITEM, col0="Item Name", col1="Item Identifier", vanilla=False)
        self.add_report(ReportType.BLOCK, col0="Block Name", col1="Block Identifier", col2="Block States", vanilla=False)
        self.add_report(ReportType.PARTICLE, col0="Particle Name", col1="Particle Identifier", vanilla=False)
        return self

    def add_report(self, report_type: ReportType, col0, **columns):
        for col_name, col_value in columns.items():
            if isinstance(col_value, tuple | set | list):
                self.dict[report_type][col0][col_name].update(col_value)
            else:
                self.dict[report_type][col0][col_name].add(str(col_value))



from typing import NewType

Color = NewType("Color", [tuple[int, int, int] | tuple[int, int, int, int] | str])
Seconds = NewType("Seconds", float)
Molang = NewType("Molang", str)
coordinate = NewType("coordinate", [float | str])
coordinates = NewType("tuple(x, y, z)", tuple[coordinate, coordinate, coordinate])
position = NewType("tuple(x, y, z)", tuple[coordinate, coordinate, coordinate])
rotation = NewType("tuple(ry,rx)", tuple[coordinate, coordinate])
level = NewType("tuple(lm,l)", tuple[float, float])
Component = NewType("Component", str)
Identifier = NewType("Identifier", str)
event = NewType("Event", str)
tick = NewType("Tick", int)
_range = NewType("[range]", str)


import logging
from datetime import datetime

import click

from ..__version__ import __version__


class Logger:
    """A class used to log messages to the console and to a log file."""

    @staticmethod
    def Red(text: str):
        return click.style(text, "red")

    @staticmethod
    def Yellow(text: str):
        return click.style(text, "yellow")

    @staticmethod
    def Green(text: str):
        return click.style(text, "green")

    @staticmethod
    def Cyan(text: str):
        return click.style(text, "cyan")

    def __init__(self) -> None:
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s - %(levelname)s - %(message)s",
            filename="anvil.log",
            filemode="w",
        )
        self.logger = logging.getLogger()

    # Console header
    def header(self):
        click.clear()
        click.echo(
            "\n".join(
                [
                    f"{self.Cyan('Anvil')} - by Yasser A. Benfoughal.",
                    f"Version {self.Cyan(__version__)}.",
                    f"Copyright © {datetime.now().year} {self.Red('StarkTMA')}.",
                    "All rights reserved.",
                    "",
                    "",
                ]
            )
        )

    # ------------------------------------------------------------------------
    # Addon Objects
    # ------------------------------------------------------------------------
    # Info
    def object_initiated(self, name: str):
        self.logger.info(f"Object initiated: {name}.")

    # Info
    def object_queued(self, name: str):
        self.logger.info(f"Object queued: {name}.")

    # Info
    def object_exported(self, name: str):
        self.logger.info(f"Object exported: {name}.")

    # Info
    def new_minecraft_build(self, old, new):
        m = f"A newer vanilla packages were found. Updating from {old} to {new}"
        self.logger.info(m)
        click.echo(m)

    # Info
    def minecraft_build_up_to_date(self):
        m = "Packages up to date"
        self.logger.info(m)
        click.echo(m)

    # Error
    def score_error(self, score):
        m = f"{[self.Red('ERROR')]}: Score objective must be 16 characters, Error at {score}"
        self.logger.error(m)
        raise ValueError(m)

    # Error
    def not_compiled(self):
        m = 'Code must be compiled before packaging, make sure to run "ANVIL.compile"'
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Info
    def packaging_zip(self):
        m = "Packaging submission .zip"
        self.logger.info(m)
        click.echo(self.Cyan(m))

    # Info
    def packaging_mcaddon(self):
        m = "Packaging .mcaddon"
        self.logger.info(m)

    # Info
    def packaging_mcworld(self):
        m = "Packaging .mcworld"
        self.logger.info(m)

    # Warn
    def file_exist_warning(self, filename):
        m = f"{filename} does not exist. This will cause validator fails."
        self.logger.warn(m)
        click.echo(self.Yellow("[WARNING]: " + m))

    # Error
    def file_exist_error(self, filename, directory):
        m = f'{filename} could not be found at "{directory}".'
        self.logger.error(m)
        raise FileNotFoundError(self.Red("[ERROR]: " + m))

    # Info
    def file_exist_info(self, filename):
        m = f"{filename} does not exist. It is optional but might have been left out unintentionally."
        self.logger.info(m)
        click.echo(self.Cyan("[INFO]: " + m))

    # Error
    def path_length_error(self, path):
        m = f"Relative file path [{path}] has [{len(path)}] characters, but cannot be more than [80] characters."
        self.logger.error(m)
        raise FileNotFoundError(self.Red("[ERROR]: " + m))

    # Error
    def component_version_error(self, component, current_version, target_version):
        m = f"The component [{component}] cannot be used with a format version below {target_version}. Current version is {current_version}."
        self.logger.error(m)
        raise FileNotFoundError(self.Red("[ERROR]: " + m))

    # Error
    def runtime_entity_error(self, entity):
        m = f"Runtime Identifier type must be a [Vanilla.Entities] type. Error at {entity}."
        self.logger.error(m)
        raise TypeError(self.Red("[ERROR]: " + m))

    # Error
    def runtime_entity_not_allowed(self, entity):
        m = f"Entity [{entity}] does not allow runtime identifier usage."
        self.logger.error(m)
        raise TypeError(self.Red("[ERROR]: " + m))

    # Error
    def unsupported_font_size(self):
        m = f"Font character_size must be a multiple of 16."
        self.logger.error(m)
        raise ValueError(self.Red("[ERROR]: " + m))

    # Error
    def unsupported_block_type(self, block):
        m = f"block must be  of a [Blocks] or [str] type. Error at {block}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def namespace_not_in_geo(self, geo_file, geo_namespace):
        m = f"The geometry file {geo_file}.geo.json doesn't contain a geometry called {geo_namespace}"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def multiple_rotations(self):
        m = f"Multiple rotation StrEnum were used. A maximum of 1 is allowed."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Warn
    def missing_state(self, target, controller, state):
        m = f'{target} {controller} is missing the animation state "{state}".'
        self.logger.warn(m)
        click.echo(self.Yellow("[WARNING]: " + m))

    # Error
    def dialogue_max_buttons(self, scene_tag, buttons_len):
        m = f"The Dialogue scene {scene_tag} has {buttons_len} buttons, The maximum allowed is 6."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def fog_start_end(self, fog_start, fog_end):
        m = f"fog_end: [{fog_end}] must be greater than fog_start: [{fog_start}]."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def unsupported_model_type(self, model):
        m = f"model must be of a [actors] or [blocks] type. Error at {model}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def no_geo_found(self, geo):
        m = f"The Geometry file {geo} does not have any geometry."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def no_anim_found(self, anim):
        m = f"The Animation file {anim} does not have any animations."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def client_type_unsupported(self, type, entity):
        m = f"{type} is not a supported Actor Type, error at {entity}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def namespaces_not_allowed(self, name):
        m = f"Identifiers are not valid name formats, error at {name}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Info
    def project_missing_config(self):
        m = f"The project require missing configuration. Please fill in the following options:"
        self.logger.info(m)
        click.echo(self.Cyan("[Info]: " + m))

    # Info
    def config_added_option(self, section: str, option: str):
        m = f"{option} was added to {section}"
        self.logger.info(m)
        click.echo(self.Cyan("[Info]: " + m))

    # Info
    def config_option_changeable(self, *options: str):
        m = f"The options {options} can be changed from the config file at any time, recompilation will be required."
        self.logger.info(m)
        click.echo(self.Cyan("[Info]: " + m))

    # Error
    @staticmethod
    def namespace_too_long(namespace):
        m = f"Namespace must be 8 characters or less. {namespace} is {len(namespace)} characters long."
        raise ValueError(Logger.Red("[ERROR]: " + m))

    # Error
    @staticmethod
    def reserved_namespace(namespace):
        m = f"{namespace} is a reserved namespace and cannot be used."
        raise ValueError(Logger.Red("[ERROR]: " + m))

    # Error
    @staticmethod
    def unique_namespace(namespace):
        m = f"Every namespace must be unique to the pack. For this pack it should be {namespace}."
        raise ValueError(Logger.Red("[ERROR]: " + m))

    # Error
    @staticmethod
    def project_name_too_long(project_name):
        m = f"Project name must be 16 characters or less. {project_name} is {len(project_name)} characters long."
        raise ValueError(Logger.Red("[ERROR]: " + m))

    # Error
    def missing_animation(self, animation_path, animation):
        m = f"{animation_path} is missing the animation {animation}"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def entity_missing_texture(self, entity):
        m = f"{entity} has no textures added."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def block_missing_texture(self, block):
        m = f"{block} has no default material instance added."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def block_missing_geometry(self, block):
        m = f"{block} has no default geometry added."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def block_state_values_out_of_range(self, block, state, values):
        m = f"Block states cannot have more than 16 values. {block} state [{state}] has {values} values."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def missing_render_controller(self, entity):
        m = f"[{entity}] is missing a render_controller"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def missing_geometry(self, entity):
        m = f"[{entity}] is missing a geometry"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def missing_texture(self, entity):
        m = f"[{entity}] is missing a texture"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def lang_error(self, text):
        m = f"Invalid localized string at {text}"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def sound_category_error(self):
        m = f"Invalid sound category"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def music_category_error(self):
        m = f"Invalid music category"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def damage_cause_error(self):
        m = f"Invalid damage cause"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # General
    def check_update(self):
        m = f"Checking for updates..."
        self.logger.info(m)

    def anvil_type_error(self, type):
        m = f"{type} Is not a valid Anvil Type."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Molang
    def molang_only(self, command):
        m = f"Molang operations only, Error at {command}"
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def digits_not_allowed(self, identifier):
        m = f"Names starting with a digit are not allowed {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def runtime_not_allowed(self, identifier):
        m = f"runtime_identifer is not allowed for packages of type : 'addon'. Error at {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def invalid_target(self, target):
        m = f"Package target must only be one of ['world', 'addon'], Error at {target}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def experimental_not_allowed(self, identifier):
        m = f"Experimental features are not allowed for packages of type : 'addon'. Error at {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Warn
    def entity_marked_as_experimental(self, identifier):
        m = f"Entity {identifier} is marked as experimental. This may be unintentional."
        self.logger.warn(m)
        click.echo(self.Yellow("[WARNING]: " + m))

    # Error
    def too_many_permutations(self, count):
        m = f"Total Block permutations exceeded 10000 ({count}). For minimal performance impact, reduce the number of permutations."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Warn
    def too_many_permutations_warn(self, count):
        m = f"Total Block permutations exceeded 10000 ({count}). For minimal performance impact, consider reducing the number of permutations."
        self.logger.warn(m)
        click.echo(self.Yellow("[WARNING]: " + m))

    # Error
    def multiple_rp_uuids(self, identifier):
        m = f"Multiple resource pack uuids found. Error at {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def multiple_bp_uuids(self, identifier):
        m = f"Multiple behavior pack uuids found. Error at {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def vanilla_override_error(self, identifier):
        m = f"Overriding vanilla features is not allowed for packages of type : 'addon'. Error at {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def invalid_score_format(self, start, score):
        m = f"Scores must start with the namespace [{start}]. Error at {score}. Consider using 'ANVIL.definitions.get_new_score' to dynamically get unique scores."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def invalid_tag_format(self, start, tag):
        m = f"Tags must start with the namespace [{start}]. Error at {tag}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def component_required(self, identifier, requirer, required):
        m = f"The component {requirer} requires the component {required} to be added to {identifier}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def recipe_max_items(self, name):
        m = f"Recipes cannot have more than 9 items. Error at {name}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def entity_not_valid_vanilla(self, name):
        m = f"{name} is not a valid vanilla entity."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def namespace_not_valid(self, namespace):
        m = f"{namespace} is not a valid namespace."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    def extend_player_rendering_not_free(self, camera_preset: str):
        m = f"Extending player component is only supported with the Free preset. Error at {camera_preset}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))

    # Error
    def rc_referenced_texture_not_in_entity(self, texture, entity):
        m = f"Referenced texture [{texture}] is not in the entity {entity}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))
    
    # Error
    def rc_referenced_geometry_not_in_entity(self, geometry, entity):
        m = f"Referenced geometry [{geometry}] is not in the entity {entity}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))
    
    # Error
    def rc_referenced_material_not_in_entity(self, material, entity):
        m = f"Referenced material [{material}] is not in the entity {entity}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))
    
    # Error
    def queryable_missing_geometry(self, geometry, entity):
        m = f"Queryable geometry {geometry} is missing from the entity {entity}."
        self.logger.error(m)
        raise RuntimeError(self.Red("[ERROR]: " + m))
    
    # Error
    def blockbench_identifier_mismatch(self, identifier: str, expected: str):
        m = f"Identifier {identifier} is not formatted correctly, expected {expected}."
        self.logger.error(m)
        raise ValueError(self.Red("[ERROR]: " + m))

    # Error
    def blockbench_animation_not_found(self, animation: str, filename:str):
        m = f"Could not find the animation [{animation}] in {filename}.bbmodel"
        self.logger.error(m)
        raise FileNotFoundError(self.Red("[ERROR]: " + m))
    
    # Error
    def blockbench_texture_not_found(self, texture: str, filename: str):
        m = f"Could not find the texture [{texture}] in {filename}.bbmodel"
        self.logger.error(m)
        raise FileNotFoundError(self.Red("[ERROR]: " + m))
    
    # Error
    def object_export_error(self, name: str, error: Exception):
        m = f"Error exporting {name}: {error}"
        self.logger.error(m)
        raise error